# -*- coding: utf-8 -*-

"""Top-level package for RHG Compute Tools."""


import pkg_resources

__author__ = """Michael Delgado"""
__email__ = "mdelgado@rhg.com"
__version__ = pkg_resources.get_distribution("rhg_compute_tools").version
